package engine

import (
	"errors"
	"fmt"
	"sync"
	"sync/atomic"
	"time"

	"github.com/KevoDB/kevo/pkg/common"
	"github.com/KevoDB/kevo/pkg/common/iterator"
	"github.com/KevoDB/kevo/pkg/config"
	"github.com/KevoDB/kevo/pkg/engine/storage"
	"github.com/KevoDB/kevo/pkg/wal"
)

var (
	// ErrEngineClosed is returned when operations are performed on a closed engine
	ErrEngineClosed = errors.New("engine is closed")
	// ErrKeyNotFound is returned when a key is not found
	ErrKeyNotFound = errors.New("key not found")
)

// EngineStats tracks statistics and metrics for the storage engine
type EngineStats struct {
	// Operation counters
	PutOps    atomic.Uint64
	GetOps    atomic.Uint64
	GetHits   atomic.Uint64
	GetMisses atomic.Uint64
	DeleteOps atomic.Uint64

	// Timing measurements
	LastPutTime    time.Time
	LastGetTime    time.Time
	LastDeleteTime time.Time

	// Performance stats
	FlushCount        atomic.Uint64
	MemTableSize      atomic.Uint64
	TotalBytesRead    atomic.Uint64
	TotalBytesWritten atomic.Uint64

	// Error tracking
	ReadErrors  atomic.Uint64
	WriteErrors atomic.Uint64

	// Transaction stats
	TxStarted   atomic.Uint64
	TxCompleted atomic.Uint64
	TxAborted   atomic.Uint64

	// WAL recovery statistics
	WALFilesRecovered   atomic.Uint64
	WALEntriesRecovered atomic.Uint64
	WALCorruptedEntries atomic.Uint64
	WALRecoveryDuration atomic.Int64

	// Mutex for timing fields
	mu sync.RWMutex
}

// Engine implements the core storage engine functionality with modular components
type Engine struct {
	// Storage components
	storage storage.Manager

	// Configuration
	cfg *config.Config

	// State management
	closed atomic.Bool

	// Statistics
	stats EngineStats

	// Concurrency control
	txLock sync.RWMutex // Lock for transaction isolation
}

// NewEngine creates a new storage engine with modular components
func NewEngine(dataDir string) (*Engine, error) {
	// Load or create the configuration
	cfg, err := config.LoadConfigFromManifest(dataDir)
	if err != nil {
		if !errors.Is(err, config.ErrManifestNotFound) {
			return nil, fmt.Errorf("failed to load configuration: %w", err)
		}
		// Create a new configuration
		cfg = config.NewDefaultConfig(dataDir)
		if err := cfg.SaveManifest(dataDir); err != nil {
			return nil, fmt.Errorf("failed to save configuration: %w", err)
		}
	}

	// Initialize the storage manager
	storageMgr, err := storage.NewDefaultStorageManager(dataDir, cfg)
	if err != nil {
		return nil, fmt.Errorf("failed to initialize storage manager: %w", err)
	}

	e := &Engine{
		cfg:     cfg,
		storage: storageMgr,
	}

	// TODO: Initialize compaction (once it's refactored)

	return e, nil
}

// Put adds a key-value pair to the database
func (e *Engine) Put(key, value []byte) error {
	// Track operation and time
	e.stats.PutOps.Add(1)

	e.stats.mu.Lock()
	e.stats.LastPutTime = time.Now()
	e.stats.mu.Unlock()

	if e.closed.Load() {
		e.stats.WriteErrors.Add(1)
		return ErrEngineClosed
	}

	// Delegate to storage manager
	_, err := e.storage.Put(key, value)
	if err != nil {
		e.stats.WriteErrors.Add(1)
		return err
	}

	// Track bytes written
	e.stats.TotalBytesWritten.Add(uint64(len(key) + len(value)))

	// Update memtable size estimate
	// TODO: This will need to be updated to get the size from the storage manager
	// e.stats.MemTableSize.Store(uint64(e.storage.GetMemTableSize()))

	return nil
}

// Get retrieves the value for the given key
func (e *Engine) Get(key []byte) ([]byte, error) {
	// Track operation and time
	e.stats.GetOps.Add(1)

	e.stats.mu.Lock()
	e.stats.LastGetTime = time.Now()
	e.stats.mu.Unlock()

	if e.closed.Load() {
		e.stats.ReadErrors.Add(1)
		return nil, ErrEngineClosed
	}

	// Track bytes read (key only at this point)
	e.stats.TotalBytesRead.Add(uint64(len(key)))

	// Delegate to storage manager
	value, found, err := e.storage.Get(key)
	if err != nil {
		if errors.Is(err, storage.ErrKeyNotFound) {
			e.stats.GetMisses.Add(1)
			return nil, ErrKeyNotFound
		}
		e.stats.ReadErrors.Add(1)
		return nil, err
	}

	if !found {
		e.stats.GetMisses.Add(1)
		return nil, ErrKeyNotFound
	}

	// Track bytes read (value part)
	e.stats.TotalBytesRead.Add(uint64(len(value)))
	e.stats.GetHits.Add(1)

	return value, nil
}

// Delete removes a key from the database
func (e *Engine) Delete(key []byte) error {
	// Track operation and time
	e.stats.DeleteOps.Add(1)

	e.stats.mu.Lock()
	e.stats.LastDeleteTime = time.Now()
	e.stats.mu.Unlock()

	if e.closed.Load() {
		e.stats.WriteErrors.Add(1)
		return ErrEngineClosed
	}

	// Delegate to storage manager
	_, err := e.storage.Delete(key)
	if err != nil {
		e.stats.WriteErrors.Add(1)
		return err
	}

	// Track bytes written (just the key for deletes)
	e.stats.TotalBytesWritten.Add(uint64(len(key)))

	return nil
}

// IsDeleted returns true if the key exists and is marked as deleted
func (e *Engine) IsDeleted(key []byte) (bool, error) {
	// This is a simplified implementation that will need to be enhanced
	// to properly check for tombstones in the storage manager

	if e.closed.Load() {
		return false, ErrEngineClosed
	}

	_, err := e.Get(key)
	if err != nil {
		if errors.Is(err, ErrKeyNotFound) {
			// The key could be not found because it was never there,
			// or because it was deleted
			// To determine if it was deleted, we'd need to check
			// if a tombstone exists for this key
			// TODO: Add support for checking tombstones directly
			return false, nil
		}
		return false, err
	}

	return false, nil
}

// FlushImMemTables flushes all immutable MemTables to disk
func (e *Engine) FlushImMemTables() error {
	if e.closed.Load() {
		return ErrEngineClosed
	}

	err := e.storage.Flush()
	if err != nil {
		return err
	}

	e.stats.FlushCount.Add(1)
	return nil
}

// ApplyBatch atomically applies a batch of operations
func (e *Engine) ApplyBatch(entries []*wal.Entry) error {
	if e.closed.Load() {
		return ErrEngineClosed
	}

	_, err := e.storage.ApplyBatch(entries)
	return err
}

// GetIterator returns an iterator over the entire keyspace
func (e *Engine) GetIterator() (iterator.Iterator, error) {
	if e.closed.Load() {
		return nil, ErrEngineClosed
	}

	return e.storage.NewIterator()
}

// GetRangeIterator returns an iterator limited to a specific key range
func (e *Engine) GetRangeIterator(startKey, endKey []byte) (iterator.Iterator, error) {
	if e.closed.Load() {
		return nil, ErrEngineClosed
	}

	return e.storage.NewRangeIterator(startKey, endKey)
}

// NOTE: TriggerCompaction and CompactRange were removed from here
// They're now defined in compaction.go

// GetStats returns the current statistics for the engine
func (e *Engine) GetStats() map[string]interface{} {
	stats := make(map[string]interface{})

	// Add operation counters
	stats["put_ops"] = e.stats.PutOps.Load()
	stats["get_ops"] = e.stats.GetOps.Load()
	stats["get_hits"] = e.stats.GetHits.Load()
	stats["get_misses"] = e.stats.GetMisses.Load()
	stats["delete_ops"] = e.stats.DeleteOps.Load()

	// Add transaction statistics
	stats["tx_started"] = e.stats.TxStarted.Load()
	stats["tx_completed"] = e.stats.TxCompleted.Load()
	stats["tx_aborted"] = e.stats.TxAborted.Load()

	// Add performance metrics
	stats["flush_count"] = e.stats.FlushCount.Load()
	stats["memtable_size"] = e.stats.MemTableSize.Load()
	stats["total_bytes_read"] = e.stats.TotalBytesRead.Load()
	stats["total_bytes_written"] = e.stats.TotalBytesWritten.Load()

	// Add error statistics
	stats["read_errors"] = e.stats.ReadErrors.Load()
	stats["write_errors"] = e.stats.WriteErrors.Load()

	// Add WAL recovery statistics
	stats["wal_files_recovered"] = e.stats.WALFilesRecovered.Load()
	stats["wal_entries_recovered"] = e.stats.WALEntriesRecovered.Load()
	stats["wal_corrupted_entries"] = e.stats.WALCorruptedEntries.Load()
	recoveryDuration := e.stats.WALRecoveryDuration.Load()
	if recoveryDuration > 0 {
		stats["wal_recovery_duration_ms"] = recoveryDuration / int64(time.Millisecond)
	}

	// Add timing information
	e.stats.mu.RLock()
	defer e.stats.mu.RUnlock()

	stats["last_put_time"] = e.stats.LastPutTime.UnixNano()
	stats["last_get_time"] = e.stats.LastGetTime.UnixNano()
	stats["last_delete_time"] = e.stats.LastDeleteTime.UnixNano()

	// TODO: Add storage statistics once they are exposed

	return stats
}

// NOTE: GetCompactionStats was removed from here
// It's now defined in compaction.go

// GetRWLock returns the transaction lock for this engine
func (e *Engine) GetRWLock() *sync.RWMutex {
	return &e.txLock
}

// Global transaction creator function
var transactionCreatorFunc common.TransactionCreator

// RegisterTransactionCreator registers a transaction creator
func RegisterTransactionCreator(creator common.TransactionCreator) {
	transactionCreatorFunc = creator
}

// BeginTransaction starts a new transaction with the given read-only flag
func (e *Engine) BeginTransaction(readOnly bool) (common.Transaction, error) {
	// Verify engine is open
	if e.closed.Load() {
		return nil, ErrEngineClosed
	}

	// Track transaction start
	e.stats.TxStarted.Add(1)

	// Check if we have a transaction creator registered
	if transactionCreatorFunc == nil {
		e.stats.WriteErrors.Add(1)
		return nil, fmt.Errorf("no transaction creator registered")
	}

	// Create a new transaction
	txn, err := transactionCreatorFunc.CreateTransaction(e, readOnly)
	if err != nil {
		e.stats.WriteErrors.Add(1)
		return nil, err
	}

	return txn, nil
}

// IncrementTxCompleted increments the completed transaction counter
func (e *Engine) IncrementTxCompleted() {
	e.stats.TxCompleted.Add(1)
}

// IncrementTxAborted increments the aborted transaction counter
func (e *Engine) IncrementTxAborted() {
	e.stats.TxAborted.Add(1)
}

// Close closes the storage engine
func (e *Engine) Close() error {
	// First set the closed flag
	wasAlreadyClosed := e.closed.Swap(true)
	if wasAlreadyClosed {
		return nil // Already closed
	}

	// Close the storage manager
	if err := e.storage.Close(); err != nil {
		return err
	}

	return nil
}