package engine

import (
	"bytes"
	"fmt"
	"os"
	"testing"

	"github.com/KevoDB/kevo/pkg/engine/storage"
	"github.com/KevoDB/kevo/pkg/wal"
)

func setupTest(t *testing.T) (string, *Engine, func()) {
	// Create a temporary directory for the test
	dir, err := os.MkdirTemp("", "engine-test-*")
	if err != nil {
		t.Fatalf("Failed to create temp dir: %v", err)
	}

	// Create the engine
	engine, err := NewEngine(dir)
	if err != nil {
		os.RemoveAll(dir)
		t.Fatalf("Failed to create engine: %v", err)
	}

	// Return cleanup function
	cleanup := func() {
		engine.Close()
		os.RemoveAll(dir)
	}

	return dir, engine, cleanup
}

func TestEngine_BasicOperations(t *testing.T) {
	_, engine, cleanup := setupTest(t)
	defer cleanup()

	// Test Put and Get
	key := []byte("test-key")
	value := []byte("test-value")

	if err := engine.Put(key, value); err != nil {
		t.Fatalf("Failed to put key-value: %v", err)
	}

	// Get the value
	result, err := engine.Get(key)
	if err != nil {
		t.Fatalf("Failed to get key: %v", err)
	}

	if !bytes.Equal(result, value) {
		t.Errorf("Got incorrect value. Expected: %s, Got: %s", value, result)
	}

	// Test Get with non-existent key
	_, err = engine.Get([]byte("non-existent"))
	if err != ErrKeyNotFound {
		t.Errorf("Expected ErrKeyNotFound for non-existent key, got: %v", err)
	}

	// Test Delete
	if err := engine.Delete(key); err != nil {
		t.Fatalf("Failed to delete key: %v", err)
	}

	// Verify key is deleted
	_, err = engine.Get(key)
	if err != ErrKeyNotFound {
		t.Errorf("Expected ErrKeyNotFound after delete, got: %v", err)
	}
}

func TestEngine_FlushAndIterate(t *testing.T) {
	_, engine, cleanup := setupTest(t)
	defer cleanup()

	// Insert some test data
	testData := []struct {
		key   string
		value string
	}{
		{"a", "1"},
		{"b", "2"},
		{"c", "3"},
		{"d", "4"},
		{"e", "5"},
	}

	for _, data := range testData {
		if err := engine.Put([]byte(data.key), []byte(data.value)); err != nil {
			t.Fatalf("Failed to put key-value: %v", err)
		}
	}

	// Force a flush to create SSTables
	if err := engine.FlushImMemTables(); err != nil {
		t.Fatalf("Failed to flush memtables: %v", err)
	}

	// Verify all keys are still accessible
	for _, data := range testData {
		value, err := engine.Get([]byte(data.key))
		if err != nil {
			t.Errorf("Failed to get key %s: %v", data.key, err)
			continue
		}
		if !bytes.Equal(value, []byte(data.value)) {
			t.Errorf("Got incorrect value for key %s. Expected: %s, Got: %s",
				data.key, data.value, string(value))
		}
	}

	// Get an iterator
	iter, err := engine.GetIterator()
	if err != nil {
		t.Fatalf("Failed to get iterator: %v", err)
	}

	// Test iterating through all keys
	iter.SeekToFirst()
	i := 0
	for iter.Valid() {
		if i >= len(testData) {
			t.Fatalf("Iterator returned more keys than expected")
		}
		if string(iter.Key()) != testData[i].key {
			t.Errorf("Iterator key mismatch. Expected: %s, Got: %s", testData[i].key, string(iter.Key()))
		}
		if string(iter.Value()) != testData[i].value {
			t.Errorf("Iterator value mismatch. Expected: %s, Got: %s", testData[i].value, string(iter.Value()))
		}
		i++
		iter.Next()
	}

	if i != len(testData) {
		t.Errorf("Iterator returned fewer keys than expected. Got: %d, Expected: %d", i, len(testData))
	}

	// Test range iterator
	rangeIter, err := engine.GetRangeIterator([]byte("b"), []byte("e"))
	if err != nil {
		t.Fatalf("Failed to get range iterator: %v", err)
	}

	expected := []struct {
		key   string
		value string
	}{
		{"b", "2"},
		{"c", "3"},
		{"d", "4"},
	}

	rangeIter.SeekToFirst()
	i = 0
	for rangeIter.Valid() {
		if i >= len(expected) {
			t.Fatalf("Range iterator returned more keys than expected")
		}
		if string(rangeIter.Key()) != expected[i].key {
			t.Errorf("Range iterator key mismatch. Expected: %s, Got: %s", expected[i].key, string(rangeIter.Key()))
		}
		if string(rangeIter.Value()) != expected[i].value {
			t.Errorf("Range iterator value mismatch. Expected: %s, Got: %s", expected[i].value, string(rangeIter.Value()))
		}
		i++
		rangeIter.Next()
	}

	if i != len(expected) {
		t.Errorf("Range iterator returned fewer keys than expected. Got: %d, Expected: %d", i, len(expected))
	}
}

func TestEngine_BatchOperations(t *testing.T) {
	_, engine, cleanup := setupTest(t)
	defer cleanup()

	// Create a batch of operations
	batch := make([]*wal.Entry, 0, 3)
	batch = append(batch, &wal.Entry{Type: wal.OpTypePut, Key: []byte("key1"), Value: []byte("value1")})
	batch = append(batch, &wal.Entry{Type: wal.OpTypePut, Key: []byte("key2"), Value: []byte("value2")})
	batch = append(batch, &wal.Entry{Type: wal.OpTypeDelete, Key: []byte("key1"), Value: nil})

	// Apply the batch
	err := engine.ApplyBatch(batch)
	if err != nil {
		t.Fatalf("Failed to apply batch: %v", err)
	}

	// Verify the operations were applied
	// key1 should be deleted
	_, err = engine.Get([]byte("key1"))
	if err != ErrKeyNotFound {
		t.Errorf("Expected key1 to be deleted, got: %v", err)
	}

	// key2 should exist
	value, err := engine.Get([]byte("key2"))
	if err != nil {
		t.Errorf("Failed to get key2: %v", err)
	} else if !bytes.Equal(value, []byte("value2")) {
		t.Errorf("Got incorrect value for key2. Expected: %s, Got: %s", "value2", string(value))
	}
}

func TestEngine_Stats(t *testing.T) {
	_, engine, cleanup := setupTest(t)
	defer cleanup()

	// Perform some operations to generate stats
	engine.Put([]byte("key1"), []byte("value1"))
	engine.Get([]byte("key1"))
	engine.Get([]byte("nonexistent"))
	engine.Delete([]byte("key1"))

	// Get stats
	stats := engine.GetStats()

	// Check basic operation counters
	if stats["put_ops"] != uint64(1) {
		t.Errorf("Expected 1 put operation, got: %v", stats["put_ops"])
	}
	if stats["get_ops"] != uint64(2) {
		t.Errorf("Expected 2 get operations, got: %v", stats["get_ops"])
	}
	if stats["delete_ops"] != uint64(1) {
		t.Errorf("Expected 1 delete operation, got: %v", stats["delete_ops"])
	}
	if stats["get_hits"] != uint64(1) {
		t.Errorf("Expected 1 get hit, got: %v", stats["get_hits"])
	}
	if stats["get_misses"] != uint64(1) {
		t.Errorf("Expected 1 get miss, got: %v", stats["get_misses"])
	}
}

// debugStorageState prints debugging information about the storage state
func debugStorageState(t *testing.T, engine *Engine) {
	if storageManager, ok := engine.storage.(*storage.DefaultStorageManager); ok {
		t.Log("Contents of active memtable:")
		memiter := storageManager.GetMemTableManager().GetActiveMemTable().NewIterator()
		for memiter.SeekToFirst(); memiter.Valid(); memiter.Next() {
			t.Logf("Key: %s, IsTombstone: %v", string(memiter.Key()), memiter.IsTombstone())
		}
		
		// Debug: Check for key-0 in all memtables
		for i, memTable := range storageManager.GetMemTableManager().GetMemTables() {
			value, found := memTable.Get([]byte("key-0"))
			t.Logf("MemTable %d: key-0 found=%v, value=%v", i, found, value)
		}
		
		// Debug: Check for key-0 in SST manager
		value, found, _ := storageManager.GetPersistentStorage().Get([]byte("key-0"))
		t.Logf("PersistentStorage: key-0 found=%v, value=%v", found, value)
		
		// Debug: Check for key-0 in storage manager
		value, found, _ = storageManager.Get([]byte("key-0"))
		t.Logf("StorageManager: key-0 found=%v, value=%v", found, value)
	}
}

func TestEngine_MultipleOperations(t *testing.T) {
	_, engine, cleanup := setupTest(t)
	defer cleanup()

	// Add many keys
	for i := 0; i < 100; i++ {
		key := []byte(fmt.Sprintf("key-%d", i))
		value := []byte(fmt.Sprintf("value-%d", i))
		if err := engine.Put(key, value); err != nil {
			t.Fatalf("Failed to put key-value: %v", err)
		}
	}

	// Check a few keys before the first flush
	for i := 0; i < 5; i += 2 {
		key := []byte(fmt.Sprintf("key-%d", i))
		value, err := engine.Get(key)
		if err != nil {
			t.Logf("Before first flush: Key key-%d not found", i)
		} else {
			t.Logf("Before first flush: Key key-%d has value %s", i, string(value))
		}
	}

	// Force a flush
	t.Log("*** FIRST FLUSH ***")
	if err := engine.FlushImMemTables(); err != nil {
		t.Fatalf("Failed to flush memtables: %v", err)
	}

	// Check same keys after flush
	for i := 0; i < 5; i += 2 {
		key := []byte(fmt.Sprintf("key-%d", i))
		value, err := engine.Get(key)
		if err != nil {
			t.Logf("After first flush: Key key-%d not found", i)
		} else {
			t.Logf("After first flush: Key key-%d has value %s", i, string(value))
		}
	}

	// Delete some keys
	t.Log("*** DELETING KEYS ***")
	for i := 0; i < 50; i += 2 {
		key := []byte(fmt.Sprintf("key-%d", i))
		if err := engine.Delete(key); err != nil {
			t.Fatalf("Failed to delete key: %v", err)
		}
	}

	// Check a few keys after deletion but before second flush
	for i := 0; i < 5; i += 2 {
		key := []byte(fmt.Sprintf("key-%d", i))
		value, err := engine.Get(key)
		if err != nil {
			t.Logf("After delete, before second flush: Key key-%d not found (OK)", i)
		} else {
			t.Logf("After delete, before second flush: Key key-%d still has value %s (PROBLEM)", i, string(value))
		}
	}

	// Force another flush
	t.Log("*** SECOND FLUSH ***")
	if err := engine.FlushImMemTables(); err != nil {
		t.Fatalf("Failed to flush memtables: %v", err)
	}

	// Check after second flush
	for i := 0; i < 5; i += 2 {
		key := []byte(fmt.Sprintf("key-%d", i))
		value, err := engine.Get(key)
		if err != nil {
			t.Logf("After second flush: Key key-%d not found (OK)", i)
		} else {
			t.Logf("After second flush: Key key-%d still has value %s (PROBLEM)", i, string(value))
		}
	}

	// Verify remaining keys
	for i := 0; i < 100; i++ {
		key := []byte(fmt.Sprintf("key-%d", i))
		value, err := engine.Get(key)
		
		if i%2 == 0 && i < 50 {
			// This key should be deleted
			if err != ErrKeyNotFound {
				t.Errorf("Expected key-%d to be deleted, but got: %v", i, err)
			}
		} else {
			// This key should exist
			expectedValue := []byte(fmt.Sprintf("value-%d", i))
			if err != nil {
				t.Errorf("Failed to get key-%d: %v", i, err)
			} else if !bytes.Equal(value, expectedValue) {
				t.Errorf("Got incorrect value for key-%d. Expected: %s, Got: %s",
					i, string(expectedValue), string(value))
			}
		}
	}
}