package storage

import (
	"testing"

	"github.com/KevoDB/kevo/pkg/memtable"
)

func TestTombstoneInMemtable(t *testing.T) {
	// Create a new memtable adapter and insert a key-value pair
	memTable := NewMemTableAdapter(memtable.NewMemTable())
	key := []byte("test-key")
	value := []byte("test-value")
	
	// Put a key-value pair
	err := memTable.Put(key, value, 1)
	if err != nil {
		t.Fatalf("Failed to put key-value: %v", err)
	}
	
	// Check that it's correctly stored
	val, found := memTable.Get(key)
	if !found {
		t.Fatalf("Key should be found after Put")
	}
	if string(val) != string(value) {
		t.Fatalf("Value mismatch: expected %s, got %s", string(value), string(val))
	}
	
	// Delete the key
	err = memTable.Delete(key, 2)
	if err != nil {
		t.Fatalf("Failed to delete key: %v", err)
	}
	
	// Check that the key is correctly marked as deleted
	val, found = memTable.Get(key)
	if !found {
		t.Fatalf("Key should still be found after Delete, but marked as tombstone")
	}
	if val != nil {
		t.Fatalf("Expected value to be nil after Delete, got %v", val)
	}
	
	// Create an iterator and check the key through it
	iter := memTable.NewIterator()
	iter.SeekToFirst()
	
	// In our memtable iterator, we expect to find entries in order of insertion,
	// so we should encounter the tombstone (seq 2) before the value (seq 1)
	found = false
	foundTombstone := false
	
	for iter.Valid() {
		if string(iter.Key()) == string(key) {
			found = true
			// Check if this is the first occurrence of the key
			if !foundTombstone {
				foundTombstone = true
				// The first occurrence should be a tombstone
				if !iter.IsTombstone() {
					t.Fatalf("Expected first occurrence of key to be a tombstone")
				}
			}
		}
		iter.Next()
	}
	
	if !found {
		t.Fatalf("Key should be found in iterator")
	}
}