package storage

import (
	"github.com/KevoDB/kevo/pkg/common/iterator"
	"github.com/KevoDB/kevo/pkg/config"
	"github.com/KevoDB/kevo/pkg/wal"
)

// Manager is the interface for managing storage operations
type Manager interface {
	// Put adds a key-value pair to the storage
	Put(key, value []byte) (uint64, error)

	// Get retrieves a value for the given key
	Get(key []byte) ([]byte, bool, error)

	// Delete marks a key as deleted
	Delete(key []byte) (uint64, error)

	// Flush flushes all pending writes to stable storage
	Flush() error

	// NewIterator returns an iterator over the entire keyspace
	NewIterator() (iterator.Iterator, error)

	// NewRangeIterator returns an iterator limited to a specific key range
	NewRangeIterator(startKey, endKey []byte) (iterator.Iterator, error)

	// ApplyBatch atomically applies a batch of operations
	ApplyBatch(entries []*wal.Entry) (uint64, error)

	// Close closes the storage manager
	Close() error
}

// MemTableManager is the interface for managing memtables
type MemTableManager interface {
	// Put adds a key-value pair to the active memtable
	Put(key, value []byte, seqNum uint64) error

	// Get retrieves a value for the given key from any memtable
	Get(key []byte) ([]byte, bool)

	// Delete marks a key as deleted in the active memtable
	Delete(key []byte, seqNum uint64) error

	// IsFlushNeeded returns true if a flush is needed
	IsFlushNeeded() bool

	// SwitchToNewMemTable switches to a new memtable and returns the previous active one
	SwitchToNewMemTable() MemTable

	// GetMemTables returns all memtables (active and immutable)
	GetMemTables() []MemTable

	// GetActiveMemTable returns the active memtable
	GetActiveMemTable() MemTable

	// SetActiveMemTable sets the active memtable
	SetActiveMemTable(MemTable)

	// TotalSize returns the total size of all memtables
	TotalSize() int64

	// Close cleans up resources
	Close() error
}

// MemTable is the interface for a single memtable
type MemTable interface {
	// Put adds a key-value pair to the memtable
	Put(key, value []byte, seqNum uint64) error

	// Get retrieves a value for the given key
	Get(key []byte) ([]byte, bool)

	// Delete marks a key as deleted
	Delete(key []byte, seqNum uint64) error

	// NewIterator returns an iterator over the memtable
	NewIterator() iterator.Iterator

	// ApproximateSize returns the approximate size of the memtable
	ApproximateSize() int64

	// IsImmutable returns true if the memtable is immutable
	IsImmutable() bool

	// SetImmutable marks the memtable as immutable
	SetImmutable()
}

// PersistentStorage is the interface for persistent storage operations
type PersistentStorage interface {
	// Flush flushes a memtable to persistent storage
	Flush(mem MemTable) error

	// Get retrieves a value for the given key from persistent storage
	Get(key []byte) ([]byte, bool, error)

	// NewIterator returns an iterator over the entire persistent storage
	NewIterator() (iterator.Iterator, error)

	// NewRangeIterator returns an iterator limited to a specific key range
	NewRangeIterator(startKey, endKey []byte) (iterator.Iterator, error)

	// Close closes the persistent storage
	Close() error
}

// WALManager is the interface for write-ahead log operations
type WALManager interface {
	// Append adds an entry to the WAL
	Append(opType byte, key, value []byte) (uint64, error)

	// AppendBatch atomically appends multiple entries to the WAL
	AppendBatch(entries []*wal.Entry) (uint64, error)

	// UpdateNextSequence updates the next sequence number
	UpdateNextSequence(seqNum uint64)

	// Rotate creates a new WAL file and closes the old one
	Rotate() error

	// Close closes the WAL
	Close() error
}

// CompactionManager is the interface for managing compaction operations
type CompactionManager interface {
	// Start starts the compaction manager
	Start() error

	// Stop stops the compaction manager
	Stop() error

	// TriggerCompaction forces a compaction cycle
	TriggerCompaction() error

	// CompactRange forces compaction on a specific key range
	CompactRange(startKey, endKey []byte) error

	// TrackTombstone tracks a tombstone for potential cleanup
	TrackTombstone(key []byte)

	// ForcePreserveTombstone forces a tombstone to be preserved
	ForcePreserveTombstone(key []byte)

	// GetCompactionStats returns statistics about the compaction state
	GetCompactionStats() map[string]interface{}
}

// ConfigProvider is the interface for accessing configuration
type ConfigProvider interface {
	// GetConfig returns the current configuration
	GetConfig() *config.Config
}