package storage

import (
	"bytes"
	"sync"

	"github.com/KevoDB/kevo/pkg/common/iterator"
)

// EmptyIterator is an iterator that contains no entries
type EmptyIterator struct{}

func (e *EmptyIterator) SeekToFirst()            {}
func (e *EmptyIterator) SeekToLast()             {}
func (e *EmptyIterator) Seek(target []byte) bool { return false }
func (e *EmptyIterator) Next() bool              { return false }
func (e *EmptyIterator) Key() []byte             { return nil }
func (e *EmptyIterator) Value() []byte           { return nil }
func (e *EmptyIterator) Valid() bool             { return false }
func (e *EmptyIterator) IsTombstone() bool       { return false }

// BoundedIterator wraps an iterator and limits it to a specific range
type BoundedIterator struct {
	iterator.Iterator
	start []byte
	end   []byte
}

// NewBoundedIterator creates a new bounded iterator
func NewBoundedIterator(iter iterator.Iterator, start, end []byte) *BoundedIterator {
	b := &BoundedIterator{
		Iterator: iter,
	}

	// Make copies of the bounds to avoid external modification
	if start != nil {
		b.start = make([]byte, len(start))
		copy(b.start, start)
	}

	if end != nil {
		b.end = make([]byte, len(end))
		copy(b.end, end)
	}

	return b
}

func (b *BoundedIterator) SeekToFirst() {
	if b.start != nil {
		// If we have a start bound, seek to it
		b.Iterator.Seek(b.start)
	} else {
		// Otherwise seek to the first key
		b.Iterator.SeekToFirst()
	}
	b.checkBounds()
}

func (b *BoundedIterator) SeekToLast() {
	if b.end != nil {
		// If we have an end bound, seek to it
		b.Iterator.Seek(b.end)

		// If we landed exactly at the end bound, back up one
		if b.Iterator.Valid() && bytes.Equal(b.Iterator.Key(), b.end) {
			// We need to back up because end is exclusive
			// This is inefficient but correct
			b.Iterator.SeekToFirst()

			// Scan to find the last key before the end bound
			var lastKey []byte
			for b.Iterator.Valid() && bytes.Compare(b.Iterator.Key(), b.end) < 0 {
				lastKey = b.Iterator.Key()
				b.Iterator.Next()
			}

			if lastKey != nil {
				b.Iterator.Seek(lastKey)
			} else {
				// No keys before the end bound
				b.Iterator.SeekToFirst()
				// This will be marked invalid by checkBounds
			}
		}
	} else {
		// No end bound, seek to the last key
		b.Iterator.SeekToLast()
	}

	// Verify we're within bounds
	b.checkBounds()
}

func (b *BoundedIterator) Seek(target []byte) bool {
	// If target is before start bound, use start bound instead
	if b.start != nil && bytes.Compare(target, b.start) < 0 {
		target = b.start
	}

	// If target is at or after end bound, the seek will fail
	if b.end != nil && bytes.Compare(target, b.end) >= 0 {
		return false
	}

	if b.Iterator.Seek(target) {
		return b.checkBounds()
	}
	return false
}

func (b *BoundedIterator) Next() bool {
	// First check if we're already at or beyond the end boundary
	if !b.checkBounds() {
		return false
	}

	// Then try to advance
	if !b.Iterator.Next() {
		return false
	}

	// Check if the new position is within bounds
	return b.checkBounds()
}

func (b *BoundedIterator) Valid() bool {
	return b.Iterator.Valid() && b.checkBounds()
}

func (b *BoundedIterator) Key() []byte {
	if !b.Valid() {
		return nil
	}
	return b.Iterator.Key()
}

func (b *BoundedIterator) Value() []byte {
	if !b.Valid() {
		return nil
	}
	return b.Iterator.Value()
}

func (b *BoundedIterator) IsTombstone() bool {
	if !b.Valid() {
		return false
	}
	return b.Iterator.IsTombstone()
}

func (b *BoundedIterator) checkBounds() bool {
	if !b.Iterator.Valid() {
		return false
	}

	// Check if the current key is before the start bound
	if b.start != nil && bytes.Compare(b.Iterator.Key(), b.start) < 0 {
		return false
	}

	// Check if the current key is beyond the end bound
	if b.end != nil && bytes.Compare(b.Iterator.Key(), b.end) >= 0 {
		return false
	}

	return true
}

// MergingIterator merges multiple iterators into a single sorted view
type MergingIterator struct {
	// List of child iterators
	iterators []iterator.Iterator
	
	// Current iterator state
	current   int // Index of the current iterator in use
	key       []byte
	value     []byte
	valid     bool
	isTombstone bool
	
	// For concurrency control
	mu        sync.Mutex
}

// NewMergingIterator creates a new merged iterator from the given iterators
// The iterators should be provided in order from newest to oldest
// (typically memtable iterators followed by sstable iterators)
func NewMergingIterator(iterators []iterator.Iterator) *MergingIterator {
	// Filter out any invalid iterators
	validIterators := make([]iterator.Iterator, 0, len(iterators))
	for _, iter := range iterators {
		if iter != nil {
			validIterators = append(validIterators, iter)
		}
	}
	
	return &MergingIterator{
		iterators: validIterators,
		current:   -1,
		valid:     false,
	}
}

func (m *MergingIterator) SeekToFirst() {
	m.mu.Lock()
	defer m.mu.Unlock()

	// Position all iterators at their first key
	for _, iter := range m.iterators {
		iter.SeekToFirst()
	}

	// Find the smallest key
	m.findSmallestKey()
}

func (m *MergingIterator) SeekToLast() {
	m.mu.Lock()
	defer m.mu.Unlock()

	// Position all iterators at their last key
	for _, iter := range m.iterators {
		iter.SeekToLast()
	}

	// Find the largest key
	m.findLargestKey()
}

func (m *MergingIterator) Seek(target []byte) bool {
	m.mu.Lock()
	defer m.mu.Unlock()

	// Position all iterators at or after the target key
	for _, iter := range m.iterators {
		iter.Seek(target)
	}

	// Find the smallest key >= target
	m.findSmallestKey()
	return m.valid
}

func (m *MergingIterator) Next() bool {
	m.mu.Lock()
	defer m.mu.Unlock()

	if !m.valid {
		return false
	}

	// Save the current key to skip duplicates
	currentKey := m.key

	// Advance the current iterator
	if m.current >= 0 && m.current < len(m.iterators) {
		m.iterators[m.current].Next()
	}

	// Find the next smallest key across all iterators
	m.findSmallestKey()
	
	// If we found the same key again from another iterator, skip it
	// Use an iterative approach instead of recursion to avoid stack overflow
	for m.valid && bytes.Equal(m.key, currentKey) {
		// Advance current iterator
		if m.current >= 0 && m.current < len(m.iterators) {
			m.iterators[m.current].Next()
		}
		
		// Find the next key
		m.findSmallestKey()
	}
	
	return m.valid
}

func (m *MergingIterator) Key() []byte {
	m.mu.Lock()
	defer m.mu.Unlock()

	if !m.valid {
		return nil
	}
	
	// Return a copy to prevent external modification
	result := make([]byte, len(m.key))
	copy(result, m.key)
	return result
}

func (m *MergingIterator) Value() []byte {
	m.mu.Lock()
	defer m.mu.Unlock()

	if !m.valid {
		return nil
	}
	
	// Return a copy to prevent external modification
	result := make([]byte, len(m.value))
	copy(result, m.value)
	return result
}

func (m *MergingIterator) Valid() bool {
	m.mu.Lock()
	defer m.mu.Unlock()

	return m.valid
}

func (m *MergingIterator) IsTombstone() bool {
	m.mu.Lock()
	defer m.mu.Unlock()

	return m.valid && m.isTombstone
}

// findSmallestKey finds the iterator with the smallest key
// This is a critical function that ensures we merge the iterators correctly
// with the right precedence (recent writes/deletes take precedence over older ones)
func (m *MergingIterator) findSmallestKey() {
	m.current = -1
	m.valid = false
	m.isTombstone = false

	var smallestKey []byte
	
	// First pass: find the smallest key
	for i, iter := range m.iterators {
		if !iter.Valid() {
			continue
		}

		if smallestKey == nil || bytes.Compare(iter.Key(), smallestKey) < 0 {
			smallestKey = iter.Key()
			m.current = i
			m.valid = true
			m.isTombstone = iter.IsTombstone()
		}
	}

	// Second pass: for the same key, newer sources (lower index) take precedence
	if m.valid {
		// Find the newest version of this key across all iterators
		// Since iterators are provided in order from newest to oldest,
		// we start from index 0 and override if we find a match
		
		// For any given key, we want to select the newest entry, which is the one from
		// the iterator with the lowest index (newest source)
		for i := 0; i < len(m.iterators); i++ {
			iter := m.iterators[i]
			if !iter.Valid() {
				continue
			}
			
			// If this iterator has the same key, use it and stop looking
			if bytes.Equal(iter.Key(), smallestKey) {
				// We've found a match in a newer source - update our selection
				m.current = i
				m.isTombstone = iter.IsTombstone()
				
				// Since we're iterating from newest to oldest (index 0 and up),
				// the first match we find is always from the newest source.
				// No need to look further - this is always the correct entry.
				break
			}
		}
	}

	if m.valid {
		// Store the current key and value
		currentIter := m.iterators[m.current]
		m.key = make([]byte, len(currentIter.Key()))
		copy(m.key, currentIter.Key())

		m.value = make([]byte, len(currentIter.Value()))
		copy(m.value, currentIter.Value())
		
		// If this is a tombstone, we need to check if it should be visible
		if m.isTombstone {
			// Tombstones delete all older entries with the same key
			// But if this key only exists as a tombstone and never had a real value,
			// we should skip it entirely by finding the next valid key
			hasRealValue := false
			
			// Look in older iterators for a real value for this key
			for i := m.current + 1; i < len(m.iterators); i++ {
				iter := m.iterators[i]
				if !iter.Valid() {
					continue
				}
				
				if bytes.Equal(iter.Key(), m.key) && !iter.IsTombstone() {
					hasRealValue = true
					break
				}
			}
			
			// If we found a tombstone for a key that doesn't exist in older iterators,
			// we should skip it entirely by finding the next key
			if !hasRealValue {
				// Save the current key
				tombstoneKey := make([]byte, len(m.key))
				copy(tombstoneKey, m.key)
				
				// Advance all iterators that have this key
				for _, iter := range m.iterators {
					if iter.Valid() && bytes.Equal(iter.Key(), tombstoneKey) {
						iter.Next()
					}
				}
				
				// Find the next smallest key
				m.findSmallestKey()
			}
		}
	} else {
		m.key = nil
		m.value = nil
	}
}

// findLargestKey finds the iterator with the largest key
func (m *MergingIterator) findLargestKey() {
	m.current = -1
	m.valid = false
	m.isTombstone = false

	var largestKey []byte
	
	// First pass: find the largest key
	for i, iter := range m.iterators {
		if !iter.Valid() {
			continue
		}

		if largestKey == nil || bytes.Compare(iter.Key(), largestKey) > 0 {
			largestKey = iter.Key()
			m.current = i
			m.valid = true
			m.isTombstone = iter.IsTombstone()
		}
	}

	// Second pass: for the same key, newer sources take precedence
	if m.valid {
		for i := 0; i < m.current; i++ {
			iter := m.iterators[i]
			if !iter.Valid() {
				continue
			}
			
			// If we find the same key in a newer iterator, it takes precedence
			if bytes.Equal(iter.Key(), largestKey) {
				m.current = i
				m.isTombstone = iter.IsTombstone()
				break
			}
		}
	}

	if m.valid {
		// Store the current key and value
		currentIter := m.iterators[m.current]
		m.key = make([]byte, len(currentIter.Key()))
		copy(m.key, currentIter.Key())

		m.value = make([]byte, len(currentIter.Value()))
		copy(m.value, currentIter.Value())
	} else {
		m.key = nil
		m.value = nil
	}
}