package storage

import (
	"sync"
	"time"

	"github.com/KevoDB/kevo/pkg/compaction"
)

// TombstoneTracker provides a bridge between the storage system and the compaction system
// for tracking tombstones. When a key is deleted, it's recorded here so compaction can
// properly handle it later.
type TombstoneTracker struct {
	// The real tombstone manager from the compaction system
	delegate compaction.TombstoneManager
	
	// Local storage for when compaction system isn't available
	localTracker *compaction.TombstoneTracker
	
	// Synchronization
	mu sync.RWMutex
}

// NewTombstoneTracker creates a new tombstone tracker
func NewTombstoneTracker() *TombstoneTracker {
	// Create a default local tracker with a reasonable retention period
	// This will be used if no compaction system is provided
	localTracker := compaction.NewTombstoneTracker(24 * time.Hour)
	
	return &TombstoneTracker{
		localTracker: localTracker,
		delegate:     localTracker, // Default to using the local tracker
	}
}

// SetDelegate sets the delegate tombstone manager from the compaction system
func (t *TombstoneTracker) SetDelegate(delegate compaction.TombstoneManager) {
	if delegate == nil {
		return
	}
	
	t.mu.Lock()
	defer t.mu.Unlock()
	
	// Transfer any tombstones from the local tracker to the delegate
	// This ensures no tombstones are lost when the compaction system is initialized
	// after the storage system has already processed some deletions
	if t.localTracker != nil && t.localTracker != delegate {
		// We could implement this transfer in the future if needed
	}
	
	t.delegate = delegate
}

// AddTombstone records a key deletion
func (t *TombstoneTracker) AddTombstone(key []byte) {
	t.mu.RLock()
	defer t.mu.RUnlock()
	
	t.delegate.AddTombstone(key)
}

// ForcePreserveTombstone marks a tombstone to be preserved indefinitely
func (t *TombstoneTracker) ForcePreserveTombstone(key []byte) {
	t.mu.RLock()
	defer t.mu.RUnlock()
	
	t.delegate.ForcePreserveTombstone(key)
}

// ShouldKeepTombstone checks if a tombstone should be preserved during compaction
func (t *TombstoneTracker) ShouldKeepTombstone(key []byte) bool {
	t.mu.RLock()
	defer t.mu.RUnlock()
	
	return t.delegate.ShouldKeepTombstone(key)
}

// RemoveTombstone removes a tombstone entry for a key
// This should be called when a previously deleted key is re-inserted
func (t *TombstoneTracker) RemoveTombstone(key []byte) {
	t.mu.Lock()
	defer t.mu.Unlock()
	
	// If local tracker, handle it directly
	if tt, ok := t.delegate.(*compaction.TombstoneTracker); ok {
		// Remove from deletions map and preserveForever map
		tt.RemoveTombstone(key)
	}
}

// GetDelegate returns the delegate tombstone manager
func (t *TombstoneTracker) GetDelegate() compaction.TombstoneManager {
	t.mu.RLock()
	defer t.mu.RUnlock()
	
	return t.delegate
}