package memtable

import (
	"github.com/KevoDB/kevo/pkg/config"
	"testing"
)

func TestMemTable_Tombstone(t *testing.T) {
	// Create a new memtable
	memTable := NewMemTable()

	// Add a key-value pair
	key := []byte("test-key")
	value := []byte("test-value")
	seqNum := uint64(1)
	memTable.Put(key, value, seqNum)

	// Verify the key exists
	result, found := memTable.Get(key)
	if !found {
		t.Fatalf("Key should exist after Put, but not found")
	}
	if string(result) != string(value) {
		t.Errorf("Expected value %s, got %s", string(value), string(result))
	}

	// Delete the key
	memTable.Delete(key, seqNum+1)

	// Verify it's marked as deleted
	result, found = memTable.Get(key)
	if !found {
		t.Errorf("Key should still exist after Delete, but not found")
	}
	if result != nil {
		t.Errorf("Value should be nil for tombstone, got %v", result)
	}

	// Verify the skiplist entry is marked as a tombstone
	entry := memTable.skipList.Find(key)
	if entry == nil {
		t.Fatalf("Entry not found in skiplist")
	}
	if entry.valueType != TypeDeletion {
		t.Errorf("Expected valueType to be TypeDeletion (%d), got %d", TypeDeletion, entry.valueType)
	}
	if entry.value != nil {
		t.Errorf("Expected value to be nil for tombstone, got %v", entry.value)
	}
}

func TestMemTablePool_Tombstone(t *testing.T) {
	// Create a config for the pool
	cfg := config.NewDefaultConfig(".")
	
	// Create a pool 
	pool := NewMemTablePool(cfg)

	// Add a key-value pair
	key := []byte("test-key")
	value := []byte("test-value")
	seqNum := uint64(1)
	pool.Put(key, value, seqNum)

	// Verify the key exists
	result, found := pool.Get(key)
	if !found {
		t.Fatalf("Key should exist after Put, but not found")
	}
	if string(result) != string(value) {
		t.Errorf("Expected value %s, got %s", string(value), string(result))
	}

	// Delete the key
	pool.Delete(key, seqNum+1)

	// Verify it's marked as deleted
	result, found = pool.Get(key)
	if !found {
		t.Errorf("Key should still exist after Delete, but not found")
	}
	if result != nil {
		t.Errorf("Value should be nil for tombstone, got %v", result)
	}

	// Add another key with the same name but a higher sequence number
	// This should NOT override the tombstone
	pool.Put(key, []byte("new value"), seqNum+2)

	// Verify the key is still deleted (tombstone has higher priority)
	result, found = pool.Get(key)
	if !found {
		t.Errorf("Key should exist after Put with higher seqNum, but not found")
	}
	if result == nil {
		t.Errorf("Expected value to be restored, but still got a tombstone")
	}
}